/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "misc.h"
#include "ss.h"
#include "regs.h"
#include "memory.h"
#include "loader.h"
#include "syscall.h"
#include "sim.h"

/* track number of insn and refs */
static SS_COUNTER_TYPE num_insn = 0;
static SS_COUNTER_TYPE num_refs = 0;

/* simulator-specific options */
char *sim_optstring = "";

void
sim_options(int argc, char **argv)
{
  char c;

  /* parse options */
  getopt_init();
  while ((c = getopt_next(argc, argv, sim_optstring)) != EOF)
    {
      switch (c) {
      }
    }
}

void
sim_config(FILE *stream)
{
}

/* set to non-zero when simulator should dump statistics */
int sim_dump_stats = FALSE;

void
sim_stats(FILE *stream)
{
  extern time_t start_time;
  time_t elapsed = MAX(time((time_t *)NULL) - start_time, 1);

  info("sim: executed %.0f instructions", (double)num_insn);
  info("sim: accessed memory %.0f times", (double)num_refs);
  info("sim: simulation time: %s (%f insts/sec)",
       elapsed_time(elapsed), (double)num_insn/(double)elapsed);
}

/* configure the execution engine */
#define SET_NPC(EXPR)		(next_PC = (EXPR))
#define CPC			(regs_PC)

#define GPR(N)			(regs_R[N])
#define SET_GPR(N,EXPR)		(regs_R[N] = (EXPR))

/* register order needed for the sparcs, knix the `^1' for mips */
#define FPR_L(N)		(regs_F.l[(N)])
#define SET_FPR_L(N,EXPR)	(regs_F.l[(N)] = (EXPR))
#define FPR_F(N)		(regs_F.f[(N)])
#define SET_FPR_F(N,EXPR)	(regs_F.f[(N)] = (EXPR))
#define FPR_D(N)		(regs_F.d[(N) >> 1])
#define SET_FPR_D(N,EXPR)	(regs_F.d[(N) >> 1] = (EXPR))

#define SET_HI(EXPR)		(regs_HI = (EXPR))
#define HI			(regs_HI)
#define SET_LO(EXPR)		(regs_LO = (EXPR))
#define LO			(regs_LO)
#define FCC			(regs_FCC)
#define SET_FCC(EXPR)		(regs_FCC = (EXPR))


#define __READ_WORD(DST_T, SRC_T, SRC)					\
  ({ SRC_T temp; num_refs++;						\
     mem_read_word(Read, (SRC), &temp);					\
     ((unsigned long)((DST_T)temp)); })

#define __READ_HALF(DST_T, SRC_T, SRC)					\
  ({ SRC_T temp; num_refs++;						\
     mem_read_half(Read, (SRC), &temp);					\
     ((unsigned long)((DST_T)temp)); })

#define __READ_BYTE(DST_T, SRC_T, SRC)					\
  ({ SRC_T temp; num_refs++;						\
     mem_read_byte(Read, (SRC), &temp);					\
     ((unsigned long)((DST_T)temp)); })

#define READ_WORD(SRC)							\
  __READ_WORD(unsigned long, unsigned long, (SRC))

#define READ_UNSIGNED_HALF(SRC)						\
  __READ_HALF(unsigned long, unsigned short, (SRC))

#define READ_SIGNED_HALF(SRC)						\
  __READ_HALF(signed long, signed short, (SRC))

#define READ_UNSIGNED_BYTE(SRC)						\
  __READ_BYTE(unsigned long, unsigned char, (SRC))

#define READ_SIGNED_BYTE(SRC)						\
  __READ_BYTE(signed long, signed char, (SRC))


#define WRITE_WORD(SRC, DST)						\
  ({ unsigned long temp = (unsigned long)(SRC);				\
     num_refs++; mem_write_word(Write, (DST), &temp); })

#define WRITE_HALF(SRC, DST)						\
  ({ unsigned short temp = (unsigned short)((unsigned long)(SRC));	\
     num_refs++; mem_write_half(Write, (DST), &temp); })

#define WRITE_BYTE(SRC,DST)						\
  ({ unsigned char temp = (unsigned char)((unsigned long)(SRC));	\
     num_refs++; mem_write_byte(Write, (DST), &temp); })


#define SYSCALL(INST)		(ss_syscall(mem_access, INST))


#define DEFFU(FU,DESC)
#define DEFICLASS(ICLASS,DESC)
#define DEFINST(OP,MSK,NAME,OPFORM,RES,CLASS,O1,O2,I1,I2,I3,EXPR)
#define DEFLDST(OP,MSK,NAME,OPFORM,RES,CLASS,O1,O2,I1,I2,I3,EXPR,DIRECT)
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)
#define CONNECT(OP)
#define IMPL
#include "ss.def"
#undef DEFFU
#undef DEFICLASS
#undef DEFINST
#undef DEFLDST
#undef DEFLINK
#undef CONNECT
#undef IMPL

/* exit when this becomes non-zero */
int sim_exit_now = FALSE;

/* longjmp here when simulation is completed */
jmp_buf sim_exit_buf;

/* instruction jump table */
static void *op_jump[];

void
sim_main(void)
{
  SS_INST_TYPE inst;
  register SS_ADDR_TYPE regs_PC, next_PC;

  info("starting functional simulation");

  num_insn = 0;
  num_refs = 0;

  regs_PC = ld_prog_entry - SS_INST_SIZE;
  next_PC = ld_prog_entry;

  /* decode all instructions */
  {
    SS_ADDR_TYPE addr;

    if (OP_MAX > 255)
      fatal("cannot do fast decoding, too many opcodes");

    fprintf(stderr, "decoding text segment...");
    for (addr=ld_text_base;
	 addr < (ld_text_base+ld_text_size);
	 addr += SS_INST_SIZE)
      {
	inst = __UNCHK_MEM_ACCESS(SS_INST_TYPE, addr);
	inst.a = SWAP_WORD(inst.a);
	inst.b = SWAP_WORD(inst.b);
	inst.a = (inst.a & ~0xff) | (unsigned long)SS_OP_ENUM(SS_OPCODE(inst));
	__UNCHK_MEM_ACCESS(SS_INST_TYPE, addr) = inst;
      }
    fprintf(stderr, "done.\n");
  }


  /* execute next instruction */
  regs_PC = next_PC;
  next_PC += 8;

  while (1)
    {
      /* maintain $r0 semantics */
      regs_R[0] = 0;

      /* keep an instruction count */
      num_insn++;

      inst = __UNCHK_MEM_ACCESS(SS_INST_TYPE, regs_PC);

      switch (SS_OPCODE(inst)) {
#define DEFFU(FU,DESC)
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR)       \
      case OP:                                                          \
        EXPR;                                                           \
        break;
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)                                 \
      case OP:                                                          \
        panic("attempted to execute a linking opcode");
#define CONNECT(OP)
#include "ss.def"
#undef DEFFU
#undef DEFINST
#undef DEFLINK
#undef CONNECT
      }

      /* execute next instruction */
      regs_PC = next_PC;
      next_PC += 8;
    }
}
